import argparse
import logging
import multiprocessing as mp
import pyfaidx
import regex
import six
import sys
import time

from motif_scraper import __script_name__, __version__, rev_comp, make_degenerate_regex, SequenceMotif, fasta_motif_scan

def main():
	#############
	# arg parse #
	#############
	parser = argparse.ArgumentParser( prog=__script_name__, epilog="%s v%s" % ( __script_name__, __version__ ) )

	# FASTA index will be created if it does not exist when pyfaidx Fasta is initialized
	parser.add_argument( 'fastaFile', help="Path to FASTA file to be scanned for degenerate sequence motifs." )
	parser.add_argument( '--motif', '-m', help="A degenerate sequence motif. Can be specified multiple times.", action='append' )
	parser.add_argument( '--motif_file', help="A file containing motifs to search file. Can be specified multiple times.", action='append' )
	parser.add_argument( '--outputFile', '-o', help="Defaults to detected_motifs.csv", default="detected_motifs.csv" )
	parser.add_argument( '--region', '-r', help="Defines region to search for sites. Use 'contig:start-end' for regions, or 'contig' for whole contig. If no regions are specified, the entire FASTA file will be searched! Starts expected to be *0 start* start and Ends *1 start*", action='append', required=False )
	parser.add_argument( '--cores', '-p', help="Run search on multiple contigs / strands simultaneously", type=int, default=1 )
	parser.add_argument( '--valid_regex', help="Query is valid regex. *WILL NOT* reverse complement. Specify sequence and strand with careful consideration.", default=False, action='store_true' )
	# molecule
	parser.add_argument( '--search_strand', '-s', help="Default searches both strands, but can be set to only one.", choices=[ '+1', '-1', 'both' ], default='both' )
	parser.add_argument( "--loglevel", choices=[ 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL' ], default='INFO' )

	args = parser.parse_args()
	
	#################
	# setup logging #
	#################
	logging.basicConfig( format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s' )
	logger = logging.getLogger( __script_name__ )
	logger.setLevel( args.loglevel )
	
	########################
	# fix up input parsing #
	########################
	if args.search_strand == "both":
		run_strands = ( '+', '-' )
	elif args.search_strand == '-1':
		run_strands = ( '-', )
	else:
		run_strands = ( '+', )
	
	if args.motif is None and args.motif_file is None:
		logger.critical( "Must specify at least one --motif or path to a --motif_file" )
		sys.exit( 1 )
		
	if args.motif is None:
		args.motif = []
	
	if args.motif_file is not None:
		for path in args.motif_file:
			with open( path, 'r' ) as MOTIF_IN:
				for line in MOTIF_IN:
					line = line.strip()
					
					if len( line ) == 0:
						continue
					elif line[0] == '#':
						continue
					args.motif.append( line.upper() )
				
	args.motif = list( set( [ seq.upper() for seq in args.motif ] ) ) # uniquify motifs in case they show up multiple times
	
	if len( args.motif ) == 0:
		logger.critical( "No motifs specified!" )
		sys.exit( 1 )
	
	######################
	# set options string #
	######################
	options = "\n%s v%s\n\nOptions\n=======\n" % ( __script_name__, __version__ )
	options += "FASTA: %s\n" % ( args.fastaFile )
	options += "Motifs to search: %s\n" % ( str( args.motif ) )
	options += "Strands of FASTA to search: %s\n" % ( str( run_strands ) )
	options += "Output file: %s\n" % ( args.outputFile )
	options += "Regions: %s\n" % ( "All contigs" if args.region is None else str( args.region ) )
	options += "Motifs are valid regex (no processing): %s\n" % ( str( args.valid_regex ) )
	options += "Processes: %s\n" % ( args.cores ) # cores
	options += "Log level: %s\n" % ( str( args.loglevel ) )
	
	logger.info( options )
		
	###################
	# Open FASTA file #
	###################
	region_list = []
	
	with pyfaidx.Fasta( args.fastaFile, as_raw=True ) as FAIDX:
		if args.region is None:
			args.region = FAIDX.keys()
		
		for reg in args.region:
			contig, start, end = pyfaidx.ucsc_split( reg )  # returns [0,1) coordinates with NoneType if not start or end

			# no need for checking for contig in FASTA file, as this is done by pyfaidx
			if start is None:
				start = 0 # pyfaidx using 0-base start. Remember for custom regions!!!
			if end is None:
				end = len( FAIDX[contig] )
			
			for curr_motif in args.motif:
				for curr_strand in run_strands:
					region_list.append( ( curr_motif, contig, start, end, curr_strand ) )
					
					logger.debug( "%s\n" % ( str( region_list[-1] ) ) )

	################
	# process data #
	#############################
	# async process the contigs #
	#############################
	work_pool = mp.Pool( processes=args.cores )
	working_data = [ work_pool.apply_async( fasta_motif_scan, args=( args.fastaFile, x, args.valid_regex ) ) for x in region_list ]
	work_output = [ x.get() for x in working_data ]
	
	##############################
	# figure order out           #
	# b/c executed asyncronously #
	##############################
	result_order = []
	
	for reg in region_list:
		index = 0
		
		for out, dict in work_output:
			if out == reg:
				logger.debug( "Returned result %s matches %s" % ( index, ' '.join( [ str(x) for x in reg ] ) ) )
				
				result_order.append( index )
				break
			else:
				index += 1
				
	if len( result_order ) != len( region_list ):
		logger.critical( "Did not find results for all requested regions" )
		sys.exit( 1 )
			
	#################
	# write outputs #
	#################
	site_count = 0
	
	with open( args.outputFile, 'w' ) as OUTFH:
		# self.contig, self.start, self.end, self.strand, self.seq, self.motif
		OUTFH.write( "Contig,Start,End,Strand,Sequence,Motif\n" )
		
		for idx in result_order:
			for site in work_output[idx][1]:
				OUTFH.write( "%s\n" % ( site ) )
				
				site_count += 1

	logger.info( "%s total sites found" % ( site_count ) )

if __name__ == "__main__":
	main()
	


#!/usr/bin/env python

"""
motif_scraper: a tool for searching for nucleic acid motifs using degenerate nucleotide queries
"""

##########
# import #
##########
import multiprocessing as mp
import pyfaidx
import regex
import six
import sys
import time

####################
# Version and name #
####################
__script_path__ = sys.argv[0]
__script_name__ = __script_path__.split('/')[-1].split('\\')[-1]
__version__ = '1.0.0'

########
# fxns #
########	
def rev_comp( seq, molecule='dna' ):
	""" DNA|RNA seq -> reverse complement
	"""
	
	if molecule == 'dna':
		nuc_dict = { "A":"T", "B":"V", "C":"G", "D":"H", "G":"C", "H":"D", "K":"M", "M":"K", "N":"N", "R":"Y", "S":"S", "T":"A", "V":"B", "W":"W", "Y":"R" }
	elif molecule == 'rna':
		nuc_dict = { "A":"U", "B":"V", "C":"G", "D":"H", "G":"C", "H":"D", "K":"M", "M":"K", "N":"N", "R":"Y", "S":"S", "U":"A", "V":"B", "W":"W", "Y":"R" }
	else:
		raise ValueError( "rev_comp requires molecule to be dna or rna" )
	
	if not isinstance( seq, six.string_types ):
		raise TypeError( "seq must be a string!" )
	
	return ''.join( [ nuc_dict[c] for c in seq.upper()[::-1] ] )

def make_degenerate_regex( motif_seq, molecule='dna' ):
	""" Degenerate sequence -> regex
	Example: NNYCGAARN -> [ACGT]{2}[CT]CGA{2}[AG][ACGT]
	"""
	
	if not isinstance( motif_seq, six.string_types ):
		raise TypeError( "motif_seq must be a string!" )
		
	if molecule == 'dna':
		degenerate_code = { "A":"A", "B":"[CGT]", "C":"C", "D":"[AGT]", "G":"G", "H":"[ACT]", "K":"[GT]", "M":"[AC]", "N":"[ACGT]", "R":"[AG]", "S":"[GC]", "T":"T", "V":"[ACG]", "W":"[AT]", "Y":"[CT]" }
	elif molecule == 'rna':
		degenerate_code = { "A":"A", "B":"[CGU]", "C":"C", "D":"[AGU]", "G":"G", "H":"[ACU]", "K":"[GU]", "M":"[AC]", "N":"[ACGU]", "R":"[AG]", "S":"[GC]", "U":"U", "V":"[ACG]", "W":"[AU]", "Y":"[CU]" }
	else:
		raise ValueError( "make_degenerate_regex requires molecule to be dna or rna" )
	
	regex_string = ''
	
	idx = 0
	
	while idx < len( motif_seq ):
		curr = motif_seq[idx]
		
		count = 1
		
		for next_idx in range( idx+1, len( motif_seq ) ):
			next = motif_seq[next_idx]
			
			if next == curr:
				count += 1
			else:
				break
		
		regex_string += degenerate_code[curr]
		
		if count > 1:
			idx = idx + count - 1
			regex_string += "{%s}" % ( count )
		
		idx += 1
	
	return regex_string
	
###########
# classes #
###########
class SequenceMotif:
	"""
	motif = the search motif
	sequence = the matching nucleic acid sequence
	contig = name of the FASTA contig where match was found
	positionStart = first base of the match*
	strand = +|- strand of match relative to FASTA sequence
	regexMatch = the regex match
	
	*positionStart is expected to be 0-base. Very important!
	"""
	
	def __init__( self, motif, contig, positionStart, strand, regexMatch, molecule='dna', regex_match_start=None, regex_end_val=None, regex_group_sequence=None ):
		if regexMatch is not None:
			if strand == '+':
				self.seq = regexMatch.group()
				self.start = positionStart + regexMatch.start() + 1
				self.end = self.start + len( self.seq ) - 1
			elif strand == '-':
				self.seq = rev_comp( regexMatch.group(), molecule )
				self.end = positionStart + regexMatch.start() + 1
				self.start = self.end + len( self.seq ) - 1
			else:
				raise ValueError( "SequenceMotif classes require strand to be either '+' or '-'" )
		else:
			if strand == '+':
				self.seq = regex_group_sequence
				self.start = positionStart + regex_match_start
				self.end = self.start + len( self.seq ) - 1
			elif strand == '-':
				self.seq = rev_comp( regex_group_sequence, molecule )
				self.end = positionStart + regex_match_start
				self.start = self.end + len( self.seq ) - 1
			else:
				raise ValueError( "SequenceMotif classes require strand to be either '+' or '-'" )
		
		self.motif = motif
		self.contig = contig
		self.strand = strand
	
	def __str__( self ):
		""" string interpolation for file writing as a CSV
		"""
		
		return "%s,%s,%s,%s,%s,%s" % ( self.contig, self.start, self.end, self.strand, self.seq, self.motif )
		   
	def __repr__( self ):
		return str( self )
	
	def __eq__( self, other ):
		if self.motif == other.motif and self.contig == other.contig and self.strand == other.strand and self.start == other.start and self.end == other.end and self.seq == other.seq:
			return True
		else:
			return False

######################
# fxns using classes #
######################
def fasta_motif_scan( fasta_fname, input_tuples, regex_ready=False, molecule='dna'  ):
	"""
	fasta_fname = string path to FASTA file
	input_tuples = tuple containing (1) motif sequence, (2) contig name, (3) start position*, (4) end position, (5) strand to search
	
	*start is expected to be 0-base, end is expected to be 1-base
	"""
	
	###################
	# validity checks #
	###################
	if not isinstance( fasta_fname, six.string_types ):
		raise TypeError( "In fasta_motif_scan, fasta_fname must be a string!" )
	elif not isinstance( molecule, six.string_types ):
		raise TypeError( "In fasta_motif_scan, molecule must be a string!" )
	elif isinstance( input_tuples, six.string_types ) or not isinstance( input_tuples, tuple ):
		raise( TypeError( "In fasta_motif_scan, input_tuples should be a tuple!" ) )
	elif not type( regex_ready ) is bool:
		raise( TypeError( "In fasta_motif_scan, regex_ready should be a bool!" ) )
	
	#########################
	# setup some local vars #
	#########################
	motif_seq, contig, start, end, strand = input_tuples
	
	if regex_ready == False:
		if strand == '+':
			regex_compiled = regex.compile( make_degenerate_regex( motif_seq ) )
		elif strand == '-':
			regex_compiled = regex.compile( make_degenerate_regex( rev_comp( motif_seq, molecule ) ) )
	else:
		if strand == '+':
			regex_compiled = regex.compile( motif_seq )
		elif strand == '-':
			regex_compiled = regex.compile( rev_comp( motif_seq, molecule ) )
	
	site_list = []
	
	with pyfaidx.Fasta( fasta_fname, as_raw=True ) as FAIDX:
		sequence = str( FAIDX[contig][start:end] ).upper()
		
		for m in regex_compiled.finditer( sequence ):
			# self, motif, contig, positionStart, strand, regexMatch, molecule='dna'
			tmp = SequenceMotif( motif_seq, contig, start, strand, m, molecule )
			site_list.append( tmp )
	
	return ( input_tuples, site_list )
	
########
# main #
########
if __name__ == "__main__":
	pass
	


from setuptools import setup

setup(
	name = "motif_scraper",
	packages = ['motif_scraper'],
	version = "1.0.0",
	description = 'Tool for finding degenerate motifs in FASTA files',
	author = "Elisha Roberson",
	author_email = 'dr.eli.roberson@gmail.com',
	url = 'https://github.com/RobersonLab/motif_scraper',
	license = 'MIT',
	classifiers=[
	"Development Status :: 5 - Production/Stable",
	"Environment :: Console",
	"Intended Audience :: Science/Research",
	"License :: OSI Approved :: MIT License",
	"Topic :: Scientific/Engineering :: Bio-Informatics",
	"Programming Language :: Python :: 2.7",
	"Programming Language :: Python :: 3",
	"Programming Language :: Python :: 3.2",
	"Programming Language :: Python :: 3.3",
	"Programming Language :: Python :: 3.4",
	"Programming Language :: Python :: 3.5"
	],
	keywords='degenerate sequence motif site search',
	install_requires = ['pyfaidx', 'regex>=2016.01.10', 'six'],
	entry_points = {'console_scripts':["motif_scraper = motif_scraper.__main__:main"]},
	test_suite = 'nose.collector',
	tests_require = ['nose']
)
